/*
簡易シグナリングサーバ
websocketを使ってクライアント用ページ間のsdp交換を行う。
ローカルPC以外からアクセスする場合は、クライアント用ページの設定で
実際のアドレスを指定する必要がある。
その場合、ファイアウォールの設定も必要になる。
use_sslをtrueにして鍵ファイルの場所を指定すればsslも対応できるが
プロトコルが割れているかつ認証なしな代物をインターネット上に公開するのは危険なので自己責任で
*/

const port_number = 14514;//websocket接続ポート番号
const use_ssl = false;//sslを使う場合はtrueにすること
//もしsslを使う場合は鍵ファイルの場所を指定すること
const ssl_server_key = 'keys/privkey.pem';
const ssl_server_crt = 'keys/cert.pem';

const fs = require('fs');
var net = require('net');
const { v4: uuidv4 } = require('uuid');
var wss;
var server;
if(use_ssl){
  const https = require('https');
  const WebSocket = require('ws');
  server = https.createServer({
    key: fs.readFileSync(ssl_server_key).toString(),
    cert: fs.readFileSync(ssl_server_crt).toString()
  });
  wss = new WebSocket.Server({
    maxPayload: 1024 * 1024,
    server 
  });
}else{
  server = require("ws").Server;
  wss = new server({port:port_number});
}

var sockets = {};
var users = {};//user-idをキーとする
var rooms = {};//部屋のurlがキーになる中身はuser-idの配列 
var dic_user2room= {};//user-idをキーとしてルーム名を値とする

function how_many_player(roomid){
  let player_num =1;
  for (let key in users) {
    if(dic_user2room[key]==roomid){
      player_num+=1;
    }
  }
  return player_num;
}

wss.on('connection', function connection(ws,req) {
  //connect時
  const ip = req.connection.remoteAddress;
  ws.ipAddress=ws._socket.remoteAddress.replace(/^.*:/g,'')
  ws.port=ws._socket.remotePort
  ws.unique=Math.floor((1 + Math.random()) * 0x10000).toString(16).substring(1)
  let tmp_id = uuidv4();
  if(tmp_id in users){
    //１回だけ重複対応
    //確率的にそれ以上は無視で
    tmp_id = uuidv4();
  }
  ws.id = uuidv4();
  //上はユーザーid

  ws.on('close', (reasonCode, description) => {
    if(ws.id in users){
      //roomsのメンバーを削る必要がある。
      let room = dic_user2room[ws.id];
      let tmp = [];
      for(let id of rooms[room]){
        if(id!= ws.id){//自分がもしあったら抜く
          tmp.push(id);
        }
      }
      rooms[room] = tmp;
      delete users[ws.id];
      delete dic_user2room[ws.id];
      delete sockets[ws.id];
    };
  })

  //receive message
  ws.on('message', function message(data,isBinary) {
    try{
      const message = isBinary ? data : data.toString();
      console.log(message);
      var obj = JSON.parse(message);
      if(obj.mode=="join"){
        //接続したら最初にjoinが来るのでwelcomeを返す
        let room = obj.room;
        sockets[ws.id] = ws;
        users[ws.id]={id:ws.id,room:room};
        //必要に応じて認証やら名前やら追加
        
        if(room in rooms){
          let tmp = [];
          for(let id of rooms[room]){
            if(id!= ws.id){//自分がもしあったら抜く
              tmp.push(id);
            }
          }
          rooms[room] = tmp;
        }else{
          rooms[room]=[];
        }
        let welcome={"mode":"welcome","you":ws.id,"members":rooms[room]}
        ws.send(JSON.stringify(welcome));
        rooms[room].push(ws.id);
        dic_user2room[ws.id] = room;
        //peerの確立は新参がofferするので
        //既存メンバーは自発的に動く必要はない。

      }else if(obj.mode=="offer"){
        let tgt = sockets[obj.target];
        tgt.send(JSON.stringify(obj));
        console.log("send offer");
      
      }else if(obj.mode=="answer"){
        let tgt = sockets[obj.target];
        tgt.send(JSON.stringify(obj));
        console.log("send answer");
      }else if(obj.mode=="candidate"){
        let tgt = sockets[obj.target];
        tgt.send(JSON.stringify(obj));
        console.log("send candidate");
      }
    }catch(e){
      console.log(e);
    }

  });

});

if(use_ssl){
  server.listen(port_number);
}


