import { 
alertAsync
,confirmAsync 
,choiceAsync
,promptAsync
,textAsync

} from "./dialogUtil";

/*
メッシュ構成のp2pクライアント
*/
/*
  基本的な使い方
  P2P_Clientをnewする。
  後はコールバックが呼ばれる
  url,callbacks,channels,rtc_config=null
  url  シグナリングサーバ
  callbacks
    onStateChange state
    onJoin id connection
    onLeave id connection
    onDataOpen id usage
    onData     id usage data  -> channelsに移行
    onDataClose id usage
  channels 名前をキーとする以下の要素の連想配列 
    type chat or  file  or transform or rpc
    onDataOpen() id 
    onData() id  data
    onDataClose() id
  rtc_config peerconnectionのconfig
*/

const sleep = msec => new Promise(resolve => setTimeout(resolve, msec));

export class P2P_Connection{
  constructor(sender,target,client){
    this.sender = sender;
    this.target = target;
    this.client = client;
    this.state = "begin";
    this.dic_channel = {};//usage to channel
    client.dic_connection[target]=this;
    this.createPeerConnection();

  };
  
  send(label,data){
    if(label in this.dic_channel){
      if(this.dic_channel[label].readyState=="open"){
        this.dic_channel[label].send(data);
      }
    }else{
      console.log("channel "+ label + " not open");
    }
  };
  
  async send_large_ext(label,_data,limit=100){
    //非同期なのでここでコピーしないと途中で書き換えられてる可能性
    const data = _data.slice(0);
    //新規チャネルを作る
    let dc = this.pc.createDataChannel(label,{});
    dc.onerror = function(error) {
      console.log('Data channel error:', error);
    };
    //送信オンリーなので何もしない
    dc.onmessage = function(evt) {
    };
    dc.onopen = async function(evt) {
      console.log('Data channel ".label." opened.');
      //オープン時に送信する
      let total = data.byteLength;
      let remain = total;
      let chunk = 50*1024;
      let begin = 0;
      let end = chunk;
      if( end > total ){
        end = total;
      };
      while(remain > 0){
        while(dc.bufferedAmount > limit){
          await sleep(20);
        };
        let buff = data.slice(begin,end);
        dc.send(buff);
        let sz = buff.byteLength;
        remain -= sz;
        begin += sz;
        end = begin+chunk;
        if( end > total ){
          end = total;
        };
      };
    };
    dc.onclose = function(evt) {
      console.log('Data channel ".label." closed.');
    };
    
  }
  
  async send_large(label,_data,limit=100){
    //大きいデータを送信する際に
    //分轄してbufferedAmountを見ながら送信する
    if(label in this.dic_channel){
      
      let channel = this.dic_channel[label];

      if(channel.readyState!="open"){
        console.log("already closed");
        return;
      }
      //非同期なのでここでコピーしないと途中で書き換えられてる可能性
      const data = _data.slice(0);

      let total = data.byteLength;
      let remain = total;
      let chunk = 50*1024;
      let begin = 0;
      let end = chunk;
      if( end > total ){
        end = total;
      };
      while(remain > 0){
        while(channel.bufferedAmount > limit){
          await sleep(20);
        };
        let buff = data.slice(begin,end);
        channel.send(buff);
        let sz = buff.byteLength;
        remain -= sz;
        begin += sz;
        end = begin+chunk;
        if( end > total ){
          end = total;
        };
      };
    }else{
      console.log("channel "+ label + " not open");
    }
    
    
    
  }
  

  async createPeerConnection() {
    let that=this;
    let sender = that.sender;
    let target = that.target;
    let client = that.client;
    this.pc = new RTCPeerConnection(client.rtc_config);

    // ICE candidate 取得時のイベントハンドラを登録
    that.pc.onicecandidate = function(evt) {
      if ( evt.candidate ) {
        console.log(evt.candidate);
        if(that.client.sig_mode=="ws" || that.client.sig_mode=="datachannel"){
          let message = {
            mode: "candidate"
            , Candidate: evt.candidate.candidate
            , sdpMLineIndex:evt.candidate.sdpMLineIndex
            , sdpMid:evt.candidate.sdpMid
            ,sender:sender
            ,target:target
          };
          that.client.signaling.send(JSON.stringify(message));
        }
        that.state = 'Collecting ICE candidates';
      } else {
        // ICE candidateの収集が完了した場合
        if(that.client.sig_mode=="client"){
          console.log('empty ice event');
          //vannilaなのでここで確定
          that.client.onOfferComplete(sender,target,peer.localDescription);
          //コールバック呼んで表示させる。
        }
      }
    };
    
    console.log("regist onconnectionstatechange",that.pc);
    that.pc.onconnectionstatechange = function(evt) {
      switch(that.pc.connectionState) {
      case "connected":
        that.state = 'connected';
        console.log("connected",that.client.sig_mode,sender,target);
        that.client._onJoin(target,that);
        if(that.client.onPeerStateChange){
          that.client.onPeerStateChange(that.target,that.state);
        }
        break;
      case "disconnected":
      case "failed":
        that.state = 'disconnected';
        that.client._onLeave(target,that);
        delete that.client.dic_connection[target];
        if(that.client.onPeerStateChange){
          that.client.onPeerStateChange(that.target,that.state);
        }
        break;
      case "closed":
        that.state = 'closed';
        delete that.client.dic_connection[target];
        if(that.client.onPeerStateChange){
          that.client.onPeerStateChange(that.target,that.state);
        }
        break;
      }
    };
    
    that.pc.ondatachannel = function(evt) {
      console.log('Data channel created:', evt);
      that.setupDataChannel(evt.channel,that.client.channels);
      //let dataChannel = evt.channel;
      //that.dc=evt.channel;
    };
    that.pc.onicegatheringstatechange = function()
    {
      console.log( "Event : ICE gathering state change" );
      console.log( "- ICE gathering state : ", that.pc.iceGatheringState );

      if( "complete" === that.pc.iceGatheringState )
      {
        let localSDP = that.pc.localDescription.sdp;
        if(that.client.sig_mode=="host" || that.client.sig_mode=="client"){
          //ここの処理は手動シグナリング用
          let evt = {
            sender:that.sender,
            target:that.target,
            sdp:localSDP,
          };
          if( that.pc.localDescription.type=="offer"){
            that.client.onOfferComplete(evt);
          }else if(that.pc.localDescription.type=="answer"){
            that.client.onAnswerComplete(evt);
          }
        }
      }
    };
    return that.pc;
  }
  setupDataChannel(dc,channels) {
    console.log("setupDataChannel",dc.label,channels);
    let that=this;
    if(! (dc.label in channels)){
      //channelsにないのは無視する
      this.client._onExtChannel(that.target,dc)
      return;
    }
    let dcinfo = channels[dc.label];
    
    
    dc.onerror = function(error) {
      console.log('Data channel error:', error);
    };
    dc.onmessage = function(evt) {
      //console.log('Data channel message:', evt.data);
      let msg = evt.data;
      if("onData" in dcinfo){
        dcinfo.onData(that.target,msg);
      }
    };
    dc.onopen = function(evt) {
      that.dic_channel[dc.label]=dc;
      if("onDataOpen" in dcinfo){
        dcinfo.onDataOpen(that.target);
      }
      console.log('Data channel opened:', evt);
    };
    dc.onclose = function(evt) {
      if("onDataClose" in dcinfo){
        dcinfo.onDataClose(that.target);
      }
      console.log('Data channel closed.');
    };

    console.log("setupDataChannel done",dc.label,channels);


  }
} 

/*
const dataChannelOptions = {
  ordered: false,
}
*/




export const pack_transform=(src)=>{
  var buffer = new ArrayBuffer(3*4+3*4+8);
  var view = new DataView(buffer);
  view.setFloat32(0,src.position.x);
  view.setFloat32(4,src.position.y);
  view.setFloat32(8,src.position.z);
  view.setFloat32(12,src.rotation.x);
  view.setFloat32(16,src.rotation.y);
  view.setFloat32(20,src.rotation.z);
  view.setFloat64(24,src.ts);
  return buffer;
}

export const unpack_transform = (buffer)=>{
  var view = new DataView(buffer);
  var res = {
    position:{x:0,y:0,z:0},
    rotation:{x:0,y:0,z:0},
    ts:0
  };
  res.position.x = view.getFloat32(0);
  res.position.y = view.getFloat32(4);
  res.position.z = view.getFloat32(8);
  res.rotation.x = view.getFloat32(12);
  res.rotation.y = view.getFloat32(16);
  res.rotation.z = view.getFloat32(20);
  res.ts = view.getFloat64(24);
  return res;
}

export const pack_position = (src)=>{
  var buffer = new ArrayBuffer(3*4);
  var view = new DataView(buffer);
  view.setFloat32(0,src.x);
  view.setFloat32(4,src.y);
  view.setFloat32(8,src.z);
  return buffer;
}

export const unpack_position=(buffer)=>{
  var view = new DataView(buffer);
  var res = {
    x:0,y:0,z:0
  };
  res.x = view.getFloat32(0);
  res.y = view.getFloat32(4);
  res.z = view.getFloat32(8);
  return res;
}



const default_rtc_config ={'iceServers': [
  { urls: "stun:stun.l.google.com:19302" },
  { urls: "stun:stun1.l.google.com:19302" },
]};

export class P2P_Client{
  constructor(url,callbacks,channels,rtc_config=null){
    let that = this;
    if(url == "as_host"){
      this.sig_mode = "host";
    }else if(url == "as_client"){
      this.sig_mode = "client";
    }else if(url == "by_datachannel"){
      this.sig_mode = "datachannel";
    }else{

      this.url = url;
      this.sig_mode = "ws";
    }
    this.my_id = "";
    this.state = "begin";
    this.dic_connection = {};//id to P2P_Connection
    this.socketReady = false;
    this.channels = channels;
    this.room = location.protocol+"/"+location.host+location.pathname;
    this.rtc_config = rtc_config?rtc_config:default_rtc_config;

    this.dic_client_id = {};//ホストモード用 クライアントのid辞書

    if("onExtChannel" in callbacks){
      this.onExtChannel = callbacks.onExtChannel;
    };

    if("onStateChange" in callbacks){
      this.onStateChange = callbacks.onStateChange;
    };
    if("onJoin" in callbacks){
      this.onJoin = callbacks.onJoin;
    };
    if("onLeave" in callbacks){
      this.onLeave = callbacks.onLeave;
    };
    if("onWelcome" in callbacks){
      this.onWelcome = callbacks.onWelcome;
    };

    if("onOfferComplete" in callbacks){
      //offerができたときの処理
      //クライアント側用
      this.onOfferComplete = callbacks.onOfferComplete;
    };

    if("onAnswerComplete" in callbacks){
      //answerができたときの処理
      //ホスト側用
      this.onAnswerComplete = callbacks.onAnswerComplete;
    };


    if("onPeerStateChange" in callbacks){
      //answerができたときの処理
      //ホスト側用
      this.onPeerStateChange = callbacks.onPeerStateChange;
    };

    this.signaling = {};//シグナリング媒体
    //wsの場合とdatachannelの場合がある
    this.signaling.on_open = (e) =>{
      console.log('signaling 接続成功');
      that.socketReady = true;
      let message= {
          mode:"join",
          room:that.room
      };
      that.signaling.send(JSON.stringify(message));
    };
    this.signaling.on_message = (e) =>{
      that.onMessage(JSON.parse(e.data));
      console.log(e.data);
    };
    //this.signaling.send


    
    if(this.sig_mode=="ws"){
      //wsを使ったシグナリング
      this.init();
    }else if(this.sig_mode=="datachannel"){
      //datachannelを使ったシグナリング
      this.init_by_datachannel(callbacks.dc);
    }else if(this.sig_mode=="host"){
      //手動シグナリング host
      this.my_id = crypto.randomUUID();
      this.init_as_host();
    }else if(this.sig_mode=="client"){
      //手動シグナリング client
      this.init_as_client();
    }




      
  }
  
  init(){
    let that = this;
    
    this.socket =  new WebSocket(this.url);
    that.signaling.send = (data) => {
      this.socket.send(data);
    };
    this.socket.addEventListener('open',that.signaling.on_open);
    that.socket.addEventListener('message',that.signaling.on_message);
  }
  
  
  init_by_datachannel(dc){
    let that=this;
    //接続済みのdcを渡す
    this.signaling.send = (data) =>{
      dc.send(data);
    }
    dc.onmessage = (evt) =>{
      //console.log('Data channel message:', evt.data);
      that.signaling.on_message(evt);
    };
    //open済みなので
    that.signaling.on_open();
  }
  
  

  init_as_host(){
    let that = this;
    //何もしない
    //クライアントのofferを受け付けた時点で手動操作で
    //onOffer_manualを呼び出す
  }

  async init_as_client(){
    let that = this;
    //id待ち
    const __offer = await promptAsync("ホストから取得したOfferを入力してください");
    const _offer = window.atob(__offer);
    const offer = JSON.parse(_offer);
    that.my_id = offer.target;
    const host_id = offer.sender;
    //こっち側はuiいらないので直接処理
    that.onAnswerComplete = async (evt) =>{
      console.log(evt);
      const _answer = JSON.stringify(evt);
      console.log(_answer);

      const __answer = window.btoa(_answer);
      console.log(__answer);

      await textAsync("Answerを取得しました",__answer,{ok:'Close',copy:"Copy"});
      
      
      
    }
    that.onOffer_manual(offer);
    //onAnswerCompleteで画面に表示する
  }


  send(id,label,data){
    //特定の相手に送信
    if(id in this.dic_connection){
      this.dic_connection[id].send(label,data);
    }else{
      console.log("no user ",id);
    }
  };

  send_large(id,label,data,limit=100){
    //特定の相手に送信
    if(id in this.dic_connection){
      this.dic_connection[id].send_large(label,data,limit);
    }else{
      console.log("no user ",id);
    }
  };
  
  send_large_ext(id,label,data,limit=100){
    //新規チャネルで特定の相手に送信
    if(id in this.dic_connection){
      this.dic_connection[id].send_large_ext(label,data,limit);
    }else{
      console.log("no user ",id);
    }
  }


  broadcast(label,data){
    //全体に送信
    for(let id in this.dic_connection){
      this.dic_connection[id].send(label,data);
    }
  };

  broadcast_large(label,data,limit = 100){
    //全体に送信
    for(let id in this.dic_connection){
      this.dic_connection[id].send_large(label,data,limit);
    }
  };
  
  onMessage(evt) {
    let that = this;
    console.log(evt);
    if (evt.mode === 'welcome') {
      if(that.onWelcome){
        that.onWelcome(evt);
      }
      that.my_id=evt.you;
      //先輩方に挨拶
      for(let target of evt.members){
        that.offer(that.my_id,target);
      }
    }else if (evt.mode === 'offer') {
        that.onOffer(evt);   
    } else if (evt.mode === 'answer') {
        that.onAnswer(evt);
    } else if (evt.mode === 'candidate' ) {
        that.onCandidate(evt);
    }
  }
  
  _onJoin(target,connection){
    if(this.onJoin){
      this.onJoin(target,connection);
    }
  }
  _onLeave(target,connection){
    if(this.onLeave){
      this.onLeave(target,connection);
    }
  }
  
  _onStateChange(st){
    this.state = st;
    if(this.onStateChange){
      this.onStateChange(st);
    }
  }


  _onExtChannel(id,channel){
    if(this.onExtChannel){
      this.onExtChannel(id,channel);
    }
  }
  
  cancel_offer(target){
    //vannilaでoffer作れなかった場合の取消用
    if(target in this.dic_connection){
      delete this.dic_connection[target];
    }
  }
  
  offer(sender,target){
    let that = this;
    console.log("offer",sender,target);
    // 新しい RTCPeerConnection を作成する
    let connection = new P2P_Connection(sender,target,this);
    let peerConnection = connection.pc;
    
    // Data channel を生成
    for(let label in that.channels){
      let options = {
      };
      let dcinfo = that.channels[label];

      if(dcinfo.type=="chat" || dcinfo.type=="file"){
        //チャットやファイル転送の場合はデフォルト
        options.ordered = true;
      }else{
        //位置同期やRPCの場合は速度優先
        options.ordered = false;
      }
      let dataChannel = peerConnection.createDataChannel(label, options);
      console.log("createDataChannel",label);
      connection.setupDataChannel(dataChannel,that.channels);
    }
    


    // Offer を生成する
    peerConnection.createOffer().then(function(sessionDescription) {
      console.log('createOffer() succeeded.');
      let localSDP = sessionDescription.sdp;
      return peerConnection.setLocalDescription(sessionDescription);
    }).then(function() {
      // setLocalDescription() が成功した場合
      // Trickle ICE ではここで SDP を相手に通知する
      // Vanilla ICE では ICE candidate が揃うのを待つ
      //シグナリングはpeerConnectionで処理している
      if(that.sig_mode=="ws" || that.sig_mode=="datachannel"){
        let message= {
            mode:"offer"
            ,sender:sender
            ,target:target
            ,sdp:peerConnection.localDescription
        };
        that.signaling.send(JSON.stringify(message));
      }
      console.log('setLocalDescription() succeeded.');
    }).catch(function(err) {
      console.error('setLocalDescription() failed.', err);
    });
    that._onStateChange("offer created");
  }
  
  addClient(){
    //ホスト用
    //画面からクライアント追加で実行
    //画面にリスト追加する。
    //生成したidをクライアントに連絡する
    const new_id = crypto.randomUUID();
    this.dic_client_id[new_id]=1;
    return new_id;
  }
  onOffer_manual(evt) {
    console.log("onOffer",evt);
    let that = this;
    let sender = evt.sender;
    let target = this.my_id;

    // Peer Connection が未生成の場合，SDP を Offer と見なす
    let offer = new RTCSessionDescription({
      type: 'offer',
      sdp: evt.sdp,
    });
    // Peer Connection を生成
    let connection = new P2P_Connection(that.my_id,evt.sender,this);
    let peerConnection = connection.pc;
    
    peerConnection.setRemoteDescription(offer).then(function() {
      console.log('setRemoteDescription() succeeded.');
      return peerConnection.createAnswer();
    }).then(function(sessionDescription ) {
      return peerConnection.setLocalDescription( sessionDescription );
    }).then(function() {
      that._onStateChange('answer created');
    }).catch(function(err) {
      console.error('setRemoteDescription() failed.', err);
    });

    peerConnection.onicecandidate = function (evt) {
      if (evt.candidate) { // ICE candidate が収集された場合
        console.log(evt.candidate);
      } else { // ICE candidateの収集が完了した場合
        console.log('empty ice event');
        //vannilaなのでここで確定
        const evt = {
          sender:that.my_id,
          target:sender,
          sdp:peerConnection.localDescription,
        };
        that.onAnswerComplete(evt);
        //コールバック呼んで表示させる。
      }
    };

  }


  onOffer(evt) {
    let remoteSDP = evt.sdp;
    console.log("onOffer",evt);
    let that = this;
    let sender = evt.sender;
    let target = evt.target;

    // Peer Connection が未生成の場合，SDP を Offer と見なす
    let offer = new RTCSessionDescription({
      type: 'offer',
      sdp: remoteSDP.sdp,
    });
    // Peer Connection を生成
    let connection = new P2P_Connection(that.my_id,evt.sender,this);
    let peerConnection = connection.pc;
    
    //trickleの場合この時点ではまだ設定できない？
    peerConnection.setRemoteDescription(offer).then(function() {
      console.log('setRemoteDescription() succeeded.');
      return peerConnection.createAnswer();
    }).then(function(sessionDescription ) {
      return peerConnection.setLocalDescription( sessionDescription );
    }).then(function() {
      that._onStateChange('answer created');
      let message= {
        mode:"answer"
        ,sender:evt.target
        ,target:evt.sender
        ,sdp:peerConnection.localDescription
      };
      that.signaling.send(JSON.stringify(message));
    }).catch(function(err) {
      console.error('setRemoteDescription() failed.', err);
    });


  }

  onAnswer_manual(evt) {
    //client用 
    let that = this;
    let remoteSDP = evt.sdp;
    let connection = that.dic_connection[evt.sender];
    let peerConnection = connection.pc;
    // Peer Connection が生成済みの場合，SDP を Answer と見なす
    let answer = new RTCSessionDescription({
      type: 'answer',
      sdp: remoteSDP.sdp,
    });
    peerConnection.setRemoteDescription(answer).then(function() {
      console.log('setRemoteDescription() succeeded.');
    }).catch(function(err) {
      console.error('setRemoteDescription() failed.', err);
    });
  }


  onAnswer(evt) {
    let that = this;
    let remoteSDP = evt.sdp;
    let connection = that.dic_connection[evt.sender];
    let peerConnection = connection.pc;
    // Peer Connection が生成済みの場合，SDP を Answer と見なす
    let answer = new RTCSessionDescription({
      type: 'answer',
      sdp: remoteSDP.sdp,
    });
    peerConnection.setRemoteDescription(answer).then(function() {
      console.log('setRemoteDescription() succeeded.');
    }).catch(function(err) {
      console.error('setRemoteDescription() failed.', err);
    });
  }
  
  onCandidate(evt) {
    let that = this;
    if(! ('sdpMLineIndex' in evt)){
      console.log("ignore candidate by null sdpMLineIndex");
      return;
    }
    let connection =that.dic_connection[evt.sender];
    let peerConnection = connection.pc;
    var candidate = new RTCIceCandidate({sdpMLineIndex:evt.sdpMLineIndex, sdpMid:evt.sdpMid, candidate:evt.Candidate});
    peerConnection.addIceCandidate(candidate);
  }
  
  
}

